
RELEASE_VERSION='0.3.960315339'
BUILD_DATE='Wed Aug  9 07:31:44 UTC 2023'
BUILD_MACHINE_INFO='Linux runner-5w-mmpfk-project-28359933-concurrent-4jv95r 5.4.235-144.344.amzn2.x86_64 #1 SMP Sun Mar 12 12:50:22 UTC 2023 x86_64 GNU/Linux'
GITLAB_CI_JOB_ID='4830934470'
GIT_COMMIT_SHA='2c3d492d2c3141cac4483c39c62a01256d9b4d8c'
GIT_BRANCH='main'
